/* Project name:
     JPEG direct load
 * Copyright:
     (c) -
 * Created by:
      Stevan A. Milinkovic
      Union University, School of Computing
      Knez Mihailova 6, 11000 Belgrade, Serbia
      e-mail: smilinkovic@raf.edu.rs
 * Revision History:
       - 2012-07-30 v 0.0.0.1 - initial
       - 2012-08-19 v 0.0.0.2 - smaller JPEG library memory footprint
 */

#include "jpeg.h"

typedef SHORT FILE;             // Place holder for ANSI file system structure

static int DISP_XS=240;
static int DISP_YS=320;

static RESULT analyze (JPGS*, UINT(*)(JPGS*, BYTE*, UINT), void*, UINT, void*);
static RESULT decode (JPGS*, UINT(*)(JPGS*, void*, RECT*), BYTE);

static sbit TFT_CS at GPIOE_ODR.B15;
BYTE Buff[3200] __attribute__ ((aligned(4)));

static volatile char autorotate = 1;
static volatile char autoscale = 1;

static void TFT_Set_Landscape(void) {
  TFT_Set_Reg(0x0002,0x0000);   // Column address start high
  TFT_Set_Reg(0x0003,0x0000);   // Column address start low
  TFT_Set_Reg(0x0004,0x0001);   // Column address end high
  TFT_Set_Reg(0x0005,0x003f);   // Column address end low

  TFT_Set_Reg(0x0006,0x0000);   // Row address start high
  TFT_Set_Reg(0x0007,0x0000);   // Row address start low
  TFT_Set_Reg(0x0008,0x0000);   // Row address end high
  TFT_Set_Reg(0x0009,0x00ef);   // Row address end low

  TFT_Set_Reg(0x16, 0x68);      // Set reading orientation
}

static void TFT_Set_Portrait(void) {
    TFT_Set_Reg(0x0002,0x0000); // Column address start high
    TFT_Set_Reg(0x0003,0x0000); // Column address start low
    TFT_Set_Reg(0x0004,0x0000); // Column address end high
    TFT_Set_Reg(0x0005,0x00ef); // Column address end low
    
    TFT_Set_Reg(0x0006,0x0000); // Row address start high
    TFT_Set_Reg(0x0007,0x0000); // Row address start low
    TFT_Set_Reg(0x0008,0x0001); // Row address end high
    TFT_Set_Reg(0x0009,0x003f); // Row address end low

    TFT_Set_Reg(0x16, 0x08);    // Set reading orientation
}


// ----------------------------
// Function to get JPEG data
// ----------------------------

static UINT input_stream (JPGS* jd, BYTE* buff, UINT n) {

  UINT i;
  UCHAR character, dummy;

  if(jd->device == 0) {
    if (buff) {           // Read n bytes from the input
      for (i=0; i < n; i++) {
        Mmc_Fat_Read(&character);
        *buff++ = character;
      }
    }
    else {                // Skip n bytes from the input
      for (i=0; i < n; i++) {
        Mmc_Fat_Read(&dummy);
      }
    }
    return n;
  }
  else {
                         // TBD for ANSI compliant file system
  }
}


// ---------------------------
// Copy image block to display
// ---------------------------

static void disp_image (int left, int right, int top, int bottom, const UINT *pat) {

  UINT pd;
  int  x, y;

  if (left > right || top > bottom) return;
  TFT_CS = 0;
  for(y=top; y<bottom+1; y++){
    TFT_Set_Address_Ptr(left,y);
    for(x=left; x<right+1; x++){
      pd = *pat++;
      TFT_Write_Data_Ptr(pd);
    }
  }
  TFT_CS = 1;
}

// ----------------------
// Output decoded block
// ----------------------

static UINT output_stream (JPGS* jd, void* bitmap, RECT* rect) {
  disp_image(rect->left, rect->right, rect->top, rect->bottom, (UINT *)bitmap);
  return 1;
}

static int load_jpg (FILE *fp, void *work, UINT workl_size) {

  JPGS jd;
  RESULT rc;
  BYTE scale;

  rc = analyze(&jd, input_stream, work, workl_size, fp);
  if (rc == OK) {
    // Determine scale factor (autoscale)
    for (scale = 0; scale < 3; scale++) {
      if ((jd.width >> scale) <= DISP_XS && (jd.height >> scale) <= DISP_YS) break;
    }
    // Start to decode the JPEG file
    rc = decode(&jd, output_stream, scale);
  }
  return rc;
}


int TFT_Ext_Jpeg (FILE *fp) {
  return (load_jpg(fp, Buff, sizeof(Buff)));
}


//-----------------------------------------------
// ZigZag-order to raster-order conversion table
//-----------------------------------------------

static const BYTE ZigZag[64] = {
         0,  1,  8, 16,  9,  2,  3, 10, 17, 24, 32, 25, 18, 11,  4,  5,
        12, 19, 26, 33, 40, 48, 41, 34, 27, 20, 13,  6,  7, 14, 21, 28,
        35, 42, 49, 56, 57, 50, 43, 36, 29, 22, 15, 23, 30, 37, 44, 51,
        58, 59, 52, 45, 38, 31, 39, 46, 53, 60, 61, 54, 47, 55, 62, 63
};


//----------------------------------------------------
// Input scale factor of Arai-Agui-Nakajima algorithm
// (scaled up 16 bits for fixed point operations)
//----------------------------------------------------

#define IPSF(n)        input_scale[n]

static const WORD input_scale[64] = {
        (WORD)(1.00000*8192), (WORD)(1.38704*8192), (WORD)(1.30656*8192), (WORD)(1.17588*8192), (WORD)(1.00000*8192), (WORD)(0.78570*8192), (WORD)(0.54120*8192), (WORD)(0.27590*8192),
        (WORD)(1.38704*8192), (WORD)(1.92388*8192), (WORD)(1.81226*8192), (WORD)(1.63099*8192), (WORD)(1.38704*8192), (WORD)(1.08979*8192), (WORD)(0.75066*8192), (WORD)(0.38268*8192),
        (WORD)(1.30656*8192), (WORD)(1.81226*8192), (WORD)(1.70711*8192), (WORD)(1.53636*8192), (WORD)(1.30656*8192), (WORD)(1.02656*8192), (WORD)(0.70711*8192), (WORD)(0.36048*8192),
        (WORD)(1.17588*8192), (WORD)(1.63099*8192), (WORD)(1.53636*8192), (WORD)(1.38268*8192), (WORD)(1.17588*8192), (WORD)(0.92388*8192), (WORD)(0.63638*8192), (WORD)(0.32442*8192),
        (WORD)(1.00000*8192), (WORD)(1.38704*8192), (WORD)(1.30656*8192), (WORD)(1.17588*8192), (WORD)(1.00000*8192), (WORD)(0.78570*8192), (WORD)(0.54120*8192), (WORD)(0.27590*8192),
        (WORD)(0.78570*8192), (WORD)(1.08979*8192), (WORD)(1.02656*8192), (WORD)(0.92388*8192), (WORD)(0.78570*8192), (WORD)(0.61732*8192), (WORD)(0.42522*8192), (WORD)(0.21677*8192),
        (WORD)(0.54120*8192), (WORD)(0.75066*8192), (WORD)(0.70711*8192), (WORD)(0.63638*8192), (WORD)(0.54120*8192), (WORD)(0.42522*8192), (WORD)(0.29290*8192), (WORD)(0.14932*8192),
        (WORD)(0.27590*8192), (WORD)(0.38268*8192), (WORD)(0.36048*8192), (WORD)(0.32442*8192), (WORD)(0.27590*8192), (WORD)(0.21678*8192), (WORD)(0.14932*8192), (WORD)(0.07612*8192)
};


//---------------------------------------------
// Conversion table for fast clipping process
//---------------------------------------------

#define BYTECLIP(v) clip_byte[(UINT)(v) & 0x3FF]

static const BYTE clip_byte[1024] = {
        0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
        32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63,
        64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87, 88, 89, 90, 91, 92, 93, 94, 95,
        96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125, 126, 127,
        128, 129, 130, 131, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 156, 157, 158, 159,
        160, 161, 162, 163, 164, 165, 166, 167, 168, 169, 170, 171, 172, 173, 174, 175, 176, 177, 178, 179, 180, 181, 182, 183, 184, 185, 186, 187, 188, 189, 190, 191,
        192, 193, 194, 195, 196, 197, 198, 199, 200, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 221, 222, 223,
        224, 225, 226, 227, 228, 229, 230, 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 251, 252, 253, 254, 255,
        255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
        255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
        255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
        255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
        255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
        255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
        255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
        255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
        0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};



//-------------------------
// Allocate a memory block
//-------------------------

static void* allocate (JPGS* jd, UINT nd) {

  char *rp = 0;
  nd = (nd + 3) & ~3;                   // Align block size to the word boundary
  if (jd->p_size >= nd) {
    jd->p_size -= nd;
    rp = (char*)jd->pool;               // Get start of available memory pool
    jd->pool = (void*)(rp + nd);        // Allocate requierd bytes
  }
  return (void*)rp;  // Return allocated memory block (NULL: no available memory)
}


// --------------------------------------------
// Create dequantization and prescaling tables
// --------------------------------------------

static UINT create_qt_tbl (JPGS* jd, const BYTE* qdata, UINT ndata) {
        
  UINT i;
  BYTE d, z;
  LONG *pb;

  while (ndata) {                            // Process all tables in the segment
    if (ndata < 65) return ERR_FMT1;         // Error: table size is unaligned
    ndata -= 65;
    d = *qdata++;                            // Get table property
    if (d & 0xF0) return ERR_FMT1;           // Error: not 8-bit resolution
    i = d & 3;                               // Get table ID
    pb = allocate(jd, 64 * sizeof (LONG));   // Allocate a memory block for the table
    if (!pb) return ERR_MEM1;                // Error: not enough memory
    jd->qttbl[i] = pb;                       // Register the table
    for (i = 0; i < 64; i++) {               // Load the table
      z = ZigZag[i];                         // Zigzag-order to raster-order conversion
      pb[z] = (LONG)((DWORD)*qdata++ * IPSF(z));   // Apply scale factor
    }
  }
  return OK;
}


// ---------------------------
// Create Huffman code tables
// ---------------------------

static UINT create_huffman_tbl (JPGS* jd, const BYTE* qdata, UINT ndata) {

  UINT i, j, b, np, cls, num;
  BYTE d, *pb, *pd;
  WORD hc, *ph;
  
  while (ndata) {                           // Process all tables in the segment
    if (ndata < 17) return ERR_FMT1;        // Error: wrong data size
    ndata -= 17;
    d = *qdata++;                           // Get table number (ID) and class
    cls = (d >> 4); num = d & 0x0F;         // class = dc(0)/ac(1), table number = 0/1
    if (d & 0xEE) return ERR_FMT1;          // Error: invalid class/number
    pb = allocate(jd, 16);                  // Allocate memory block for the bit distribution table
    if (!pb) return ERR_MEM1;               // Error: not enough memory
    jd->huffbits[num][cls] = pb;
    for (np = i = 0; i < 16; i++) {         // Load number of patterns for 1 to 16-bit code
      pb[i] = b = *qdata++;
      np += b;                              // Get sum of code words for each code
    }

    ph = allocate(jd, np * sizeof (WORD));  // Allocate memory block for the code word table
    if (!ph) return ERR_MEM1;               // Error: not enough memory
    jd->huffcode[num][cls] = ph;
    hc = 0;
    for (j = i = 0; i < 16; i++) {          // Rebuild Huffman code word table
      b = pb[i];
      while (b--) ph[j++] = hc++;
      hc <<= 1;
    }

    if (ndata < np) return ERR_FMT1;        // Error: wrong data size
    ndata -= np;
    pd = allocate(jd, np);                  // Allocate a memory block for the decoded data
    if (!pd) return ERR_MEM1;               // Error: not enough memory
    jd->huffdata[num][cls] = pd;
    for (i = 0; i < np; i++) {              // Load decoded data that corresponds to each code word
      d = *qdata++;
      if (!cls && d > 11) return ERR_FMT1;
      *pd++ = d;
    }
  }
  return OK;
}


// --------------------------
// Extract N bits from input
// --------------------------

static INT bitext (JPGS* jd, UINT nbit) {

  BYTE msk, s, *dp;
  UINT dc, v, f;

  msk = jd->dmsk; dc = jd->dctr; dp = jd->dptr;  // Bit mask, number of data available, read pointer
  s = *dp; v = f = 0;
  do {
    if (!msk) {                                  // Next byte?
      if (!dc) {                                 // No input data is available, refill input buffer
        dp = jd->inbuf;                          // Top of input buffer
        dc = jd->virtual_in(jd, dp, STREAM_BUF);
          if (!dc) return 0 - ERR_INP;           // Error: read error or wrong stream termination
        } else {
            dp++;                                // Next data
        }
          dc--;                                  // Decrement number of available bytes
          if (f) {                               // In flag sequence?
            f = 0;                               // Exit flag sequence
            if (*dp != 0) return 0 - ERR_FMT1;   // Error: unexpected flag is detected
            *dp = s = 0xFF;                      // The flag is 0xFF
          } else {
              s = *dp;                           // Get next data byte
              if (s == 0xFF) {                   // Is it start of flag sequence?
                f = 1; continue;                 // Enter flag sequence
              }
          }
            msk = 0x80;                          // Read from MSB
          }
        v <<= 1;                                 // Get a bit
        if (s & msk) v++;
        msk >>= 1;
        nbit--;
      } while (nbit);
      jd->dmsk = msk; jd->dctr = dc; jd->dptr = dp;
      return (INT)v;
}


// ----------------------------------------
// Extract Huffman decoded data from input
// ----------------------------------------

static INT huffext (JPGS* jd, const BYTE* hbits, const WORD* hcode, const BYTE* hdata) {

  BYTE msk, s, *dp;
  UINT dc, v, f, bl, nd;

  msk = jd->dmsk; dc = jd->dctr; dp = jd->dptr;    // Bit mask, number of data available, read pointer
  s = *dp; v = f = 0;
  bl = 16;                                         // Max code length
  do {
    if (!msk) {                                    // Next byte?
      if (!dc) {                                   // No input data is available, refill input buffer
        dp = jd->inbuf;                            // Top of input buffer
        dc = jd->virtual_in(jd, dp, STREAM_BUF);
        if (!dc) return 0 - ERR_INP;               // Error: read error or wrong stream termination
       } else {
         dp++;                                     // Next data
       }
       dc--;                                       // Decrement number of available bytes
       if (f) {                                    // In flag sequence?
         f = 0;                                    // Exit flag sequence
         if (*dp != 0)
           return 0 - ERR_FMT1;                    // Error: unexpected flag is detected
         *dp = s = 0xFF;                           // The flag 0xFF
       } else {
         s = *dp;                                  // Get next data byte
         if (s == 0xFF) {                          // Is it start of flag sequence?
           f = 1; continue;                        // Enter flag sequence, get trailing byte
         }
       }
       msk = 0x80;                                 // Read from MSB
    }
    v <<= 1;                                       // Get a bit
    if (s & msk) v++;
    msk >>= 1;

    for (nd = *hbits++; nd; nd--) {                // Search the code word with this bit length
      if (v == *hcode++) {                         // Match?
        jd->dmsk = msk; jd->dctr = dc; jd->dptr = dp;
        return *hdata;                             // Return the decoded data
      }
      hdata++;
    }
    bl--;
  } while (bl);
  return 0 - ERR_FMT1;                             // Error: code not found
}


// -------------------
// Apply Inverse-DCT
// -------------------

static void block_idct (LONG* src, BYTE* dst) {

  const LONG M13 = (LONG)(1.41421*4096), M2 = (LONG)(1.08239*4096), M4 = (LONG)(2.61313*4096), M5 = (LONG)(1.84776*4096);
  LONG v0, v1, v2, v3, v4, v5, v6, v7;
  LONG t10, t11, t12, t13;
  UINT i;

   // Process columns
   for (i = 0; i < 8; i++) {
     v0 = src[8 * 0];              // Get even elements
     v1 = src[8 * 2];
     v2 = src[8 * 4];
     v3 = src[8 * 6];

     t10 = v0 + v2;                // Process even elements
     t12 = v0 - v2;
     t11 = (v1 - v3) * M13 >> 12;
     v3 += v1;
     t11 -= v3;
     v0 = t10 + v3;
     v3 = t10 - v3;
     v1 = t11 + t12;
     v2 = t12 - t11;

     v4 = src[8 * 7];              // Get odd elements
     v5 = src[8 * 1];
     v6 = src[8 * 5];
     v7 = src[8 * 3];

     t10 = v5 - v4;                // Process the odd elements
     t11 = v5 + v4;
     t12 = v6 - v7;
     v7 += v6;
     v5 = (t11 - v7) * M13 >> 12;
     v7 += t11;
     t13 = (t10 + t12) * M5 >> 12;
     v4 = t13 - (t10 * M2 >> 12);
     v6 = t13 - (t12 * M4 >> 12) - v7;
     v5 -= v6;
     v4 -= v5;

     src[8 * 0] = v0 + v7;         // Write-back the transformed values
     src[8 * 7] = v0 - v7;
     src[8 * 1] = v1 + v6;
     src[8 * 6] = v1 - v6;
     src[8 * 2] = v2 + v5;
     src[8 * 5] = v2 - v5;
     src[8 * 3] = v3 + v4;
     src[8 * 4] = v3 - v4;

     src++;                        // Next column
  }

  // Process rows
  src -= 8;
  for (i = 0; i < 8; i++) {
    v0 = src[0] + (128L << 8);     // Get even elements (remove DC offset (-128))
    v1 = src[2];
    v2 = src[4];
    v3 = src[6];

    t10 = v0 + v2;                 // Process the even elements
    t12 = v0 - v2;
    t11 = (v1 - v3) * M13 >> 12;
    v3 += v1;
    t11 -= v3;
    v0 = t10 + v3;
    v3 = t10 - v3;
    v1 = t11 + t12;
    v2 = t12 - t11;

    v4 = src[7];                   // Get odd elements
    v5 = src[1];
    v6 = src[5];
    v7 = src[3];

    t10 = v5 - v4;                 // Process the odd elements
    t11 = v5 + v4;
    t12 = v6 - v7;
    v7 += v6;
    v5 = (t11 - v7) * M13 >> 12;
    v7 += t11;
    t13 = (t10 + t12) * M5 >> 12;
    v4 = t13 - (t10 * M2 >> 12);
    v6 = t13 - (t12 * M4 >> 12) - v7;
    v5 -= v6;
    v4 -= v5;
    
    // Descale the transformed values 8 bits and output
    dst[0] = BYTECLIP((v0 + v7) >> 8);
    dst[7] = BYTECLIP((v0 - v7) >> 8);
    dst[1] = BYTECLIP((v1 + v6) >> 8);
    dst[6] = BYTECLIP((v1 - v6) >> 8);
    dst[2] = BYTECLIP((v2 + v5) >> 8);
    dst[5] = BYTECLIP((v2 - v5) >> 8);
    dst[3] = BYTECLIP((v3 + v4) >> 8);
    dst[4] = BYTECLIP((v3 - v4) >> 8);
    dst += 8;

    src += 8;                      // Next row
  }
}


// ------------------------------------------------
// Load all blocks in MCU into the working buffer
// ------------------------------------------------

static RESULT mcu_load (JPGS* jd) {

  LONG *tmp = (LONG*)jd->workbuf;            // Working buffer for dequantization and IDCT
  UINT blk, nby, nbc, i, z, id, cmp;
  INT b, d, e;
  BYTE *bp;
  const BYTE *hb, *hd;
  const WORD *hc;
  const LONG *dqf;

  nby = jd->msx * jd->msy;                   // Number of Y blocks (1, 2 or 4)
  nbc = 2;                                   // Number of C blocks (2)
  bp = jd->mcubuf;                           // Pointer to the first block

  for (blk = 0; blk < nby + nbc; blk++) {
    cmp = (blk < nby) ? 0 : blk - nby + 1;   // Component number 0:Y, 1:Cb, 2:Cr
    id = cmp ? 1 : 0;                        // Huffman table ID of the component

    // Extract a DC element from input stream
    hb = jd->huffbits[id][0];                // Huffman table for the DC element
    hc = jd->huffcode[id][0];
    hd = jd->huffdata[id][0];
    b = huffext(jd, hb, hc, hd);             // Extract a Huffman coded data (bit length)
    if (b < 0) return 0 - b;                 // Error: invalid code or input
    d = jd->dcv[cmp];                        // DC value of previous block
    if (b) {                                 // If there is any difference from previous block
      e = bitext(jd, b);                     // Extract data bits
      if (e < 0) return 0 - e;               // Error: input
      b = 1 << (b - 1);                      // MSB position
      if (!(e & b)) e -= (b << 1) - 1;       // Restore sign if needed
      d += e;                                // Get current value
      jd->dcv[cmp] = (SHORT)d;               // Save current DC value for next block
    }
    dqf = jd->qttbl[jd->qtid[cmp]];          // Dequantizer table ID for this component
    tmp[0] = d * dqf[0] >> 8;                // Dequantize, apply scale factor of Arai-Agui-Nakajima algorithm and descale 8 bits

    // Extract following 63 AC elements from input stream
    for (i = 1; i < 64; i++) tmp[i] = 0;     // Clear rest of elements
    hb = jd->huffbits[id][1];                // Huffman table for the AC elements
    hc = jd->huffcode[id][1];
    hd = jd->huffdata[id][1];
    i = 1;                                   // Top of the AC elements
    do {
      b = huffext(jd, hb, hc, hd);           // Extract a Huffman coded value (zero runs and bit length)
      if (b == 0) break;                     // EOB?
      if (b < 0) return 0 - b;               // Error: invalid code or input error
      z = (UINT)b >> 4;                      // Number of leading zero elements
      if (z) {
        i += z;                              // Skip zero elements
        if (i >= 64) return ERR_FMT1;        // Too long zero run
      }
      if (b &= 0x0F) {                       // Bit length
        d = bitext(jd, b);                   // Extract data bits
        if (d < 0) return 0 - d;             // Error: input device
        b = 1 << (b - 1);                    // MSB position
        if (!(d & b)) d -= (b << 1) - 1;     // Restore negative value if needed
        z = ZigZag[i];                       // Zigzag-order to raster-order converted index
        tmp[z] = d * dqf[z] >> 8;            // Dequantize, apply scale factor of Arai-Agui-Nakajima algorithm and descale 8 bits
      }
    } while (++i < 64);                      // Next AC element

    if (autoscale && jd->scale == 3)
      *bp = (*tmp / 256) + 128;              // If scale ratio is 1/8, IDCT can be ommited and only DC element is used
    else
      block_idct(tmp, bp);                   // Apply IDCT and store the block to the MCU buffer
    bp += 64;                                // Next block
  }
  return OK;                                 // All blocks have been loaded successfully
}


// ----------------------------------------------------------------
// Output the MCU: Convert YCrCb to RGB and output it in RGB form
// ----------------------------------------------------------------

static RESULT mcu_output (JPGS* jd, UINT (*virtual_out)(JPGS*, void*, RECT*), UINT x, UINT y) {

  const INT CVACC = (sizeof (INT) > 2) ? 1024 : 128;
  UINT ix, iy, mx, my, rxx, ry;
  INT yy, cb, cr;
  BYTE *py, *pc, *rgb24;
  RECT rect;

  mx = jd->msx * 8; my = jd->msy * 8;                  // MCU size (pixel)
  rxx = (x + mx <= jd->width) ? mx : jd->width - x;    // Output rectangular size
  ry = (y + my <= jd->height) ? my : jd->height - y;
  if (autoscale) {
    rxx >>= jd->scale; ry >>= jd->scale;
    if (!rxx || !ry) return OK;                        // Skip this MCU if all pixel is to be rounded off
    x >>= jd->scale; y >>= jd->scale;
  }
  rect.left = x; rect.right = x + rxx - 1;             // Rectangular area in the frame buffer
  rect.top = y; rect.bottom = y + ry - 1;

  if (!autoscale || jd->scale != 3) {                  // Not for 1/8 scaling
    rgb24 = (BYTE*)jd->workbuf;                        // Build an RGB MCU from discrete comopnents
    for (iy = 0; iy < my; iy++) {
      pc = jd->mcubuf;
      py = pc + iy * 8;
      if (my == 16) {                                  // Double block height?
        pc += 64 * 4 + (iy >> 1) * 8;
        if (iy >= 8) py += 64;
      } else {                                         // Single block height
        pc += mx * 8 + iy * 8;
      }
      for (ix = 0; ix < mx; ix++) {
        cb = pc[0] - 128;                              // Get Cb/Cr component and restore right level
        cr = pc[64] - 128;
        if (mx == 16) {                                // Double block width?
          if (ix == 8) py += 64 - 8;                   // Jump to next block if double block heigt
          pc += ix & 1;                                // Increase chroma pointer every two pixels
        } else {                                       // Single block width
          pc++;                                        // Increase chroma pointer every pixel
        }
        yy = *py++;                                    // Get Y component

       // Convert YCbCr to RGB
       *rgb24++ = BYTECLIP(yy + ((INT)(1.402 * CVACC) * cr) / CVACC);  // R
       *rgb24++ = BYTECLIP(yy - ((INT)(0.344 * CVACC) * cb + (INT)(0.714 * CVACC) * cr) / CVACC); // G
       *rgb24++ = BYTECLIP(yy + ((INT)(1.772 * CVACC) * cb) / CVACC);  // B
      }
    }
    // Descale the MCU rectangular if needed
    if (autoscale && jd->scale) {
      UINT x, y, r, g, b, s, w, a;
      BYTE *op;

      // Get averaged RGB value of each square that will correspond to one pixel
      s = jd->scale * 2;                               // Number of shifts for averaging
      w = 1 << jd->scale;                              // Width of square
      a = (mx - w) * 3;                                // Bytes to skip for next line in the square
      op = (BYTE*)jd->workbuf;
      for (iy = 0; iy < my; iy += w) {
        for (ix = 0; ix < mx; ix += w) {
          rgb24 = (BYTE*)jd->workbuf + (iy * mx + ix) * 3;
          r = g = b = 0;
          for (y = 0; y < w; y++) {                    // Accumulate RGB value in the square
            for (x = 0; x < w; x++) {
              r += *rgb24++;
              g += *rgb24++;
              b += *rgb24++;
            }
            rgb24 += a;
          }                                            // Put the averaged RGB value as a pixel
          *op++ = (BYTE)(r >> s);
          *op++ = (BYTE)(g >> s);
          *op++ = (BYTE)(b >> s);
        }
      }
    }
  } else {      // Only for  1/8 scaling (left-top pixel in each block is the DC value of the block)
                // Build a 1/8 descaled RGB MCU from discrete comopnents
    rgb24 = (BYTE*)jd->workbuf;
    pc = jd->mcubuf + mx * my;
    cb = pc[0] - 128;                                  // Get Cb/Cr component and restore right level
    cr = pc[64] - 128;
    for (iy = 0; iy < my; iy += 8) {
      py = jd->mcubuf;
      if (iy == 8) py += 64 * 2;
      for (ix = 0; ix < mx; ix += 8) {
        yy = *py;                                      // Get Y component
        py += 64;
        // Convert YCbCr to RGB
        *rgb24++ = BYTECLIP(yy + ((INT)(1.402 * CVACC) * cr / CVACC));      // R
        *rgb24++ = BYTECLIP(yy - ((INT)(0.344 * CVACC) * cb + (INT)(0.714 * CVACC) * cr) / CVACC);  // G
        *rgb24++ = BYTECLIP(yy + ((INT)(1.772 * CVACC) * cb / CVACC));      // B
      }
    }
  }
  // Squeeze up pixel table if a part of MCU is to be truncated
  mx >>= jd->scale;
  if (rxx < mx) {
    BYTE *s, *d;
    UINT x, y;

    s = d = (BYTE*)jd->workbuf;
    for (y = 0; y < ry; y++) {
      for (x = 0; x < rxx; x++) {                      // Copy effective pixels
        *d++ = *s++;
        *d++ = *s++;
        *d++ = *s++;
      }
      s += (mx - rxx) * 3;                             // Skip truncated pixels
    }
  }
  // Convert RGB888 to RGB565
  {
    BYTE *s = (BYTE*)jd->workbuf;
    WORD w, *d = (WORD*)s;
    UINT n = rxx * ry;
    do {
      w = (*s++ & 0xF8) << 8;                          // R(5)
      w |= (*s++ & 0xFC) << 3;                         // G(6)
      w |= *s++ >> 3;                                  // B(5)
      *d++ = w;
    } while (--n);
  }
  // Output the RGB rectangular
  return virtual_out(jd, jd->workbuf, &rect) ? OK : ERR_INTR;
}


// ------------------------------------------------------------
// Analyze the JPEG image and Initialize JPEG decoding object
// ------------------------------------------------------------

static RESULT analyze (JPGS* jd, UINT (*virtual_in)(JPGS*, BYTE*, UINT), void* pool, UINT p_size, void* dev) {

  BYTE *seg, b;
  WORD marker;
  DWORD ofs;
  UINT n, i, j, len;
  RESULT rc;

  if (!pool) return ERR_PAR;
  jd->pool = pool;                                   // Working memroy
  jd->p_size = p_size;                               // Size of given working memory
  jd->virtual_in = virtual_in;                       // Stream input function
  jd->device = dev;                                  // I/O device identifier
  jd->nrst = 0;                                      // No restart interval (default, i.e. not used)

  for (i = 0; i < 2; i++) {                          // Null pointers
    for (j = 0; j < 2; j++) {
      jd->huffbits[i][j] = 0;
      jd->huffcode[i][j] = 0;
    jd->huffdata[i][j] = 0;
    }
  }
  for (i = 0; i < 4; i++) jd->qttbl[i] = 0;
  jd->inbuf = seg = allocate(jd, STREAM_BUF);           // Allocate stream input buffer
  if (!seg) return ERR_MEM1;
  if (jd->virtual_in(jd, seg, 2) != 2) return ERR_INP;  // Check SOI marker
  if (LDB_WORD(seg) != 0xFFD8) return ERR_FMT1;         // Error: SOI is not detected
  ofs = 2;

  for (;;) {
  // Get a JPEG marker
    if (jd->virtual_in(jd, seg, 4) != 4) return ERR_INP;
    marker = LDB_WORD(seg);                         // Marker
    len = LDB_WORD(seg + 2);                        // Length field
    if (len <= 2 || (marker >> 8) != 0xFF) return ERR_FMT1;
    len -= 2;                                       // Content size excluding length field
    ofs += 4 + len;                                 // Number of bytes loaded

    switch (marker & 0xFF) {
      case 0xC0:                                    // SOF0 (baseline JPEG)
                                                    // Load segment data
        if (len > STREAM_BUF) return ERR_MEM2;
        if (jd->virtual_in(jd, seg, len) != len) return ERR_INP;
        jd->width = LDB_WORD(seg+3);                // Image width in pixels
        jd->height = LDB_WORD(seg+1);               // Image height in pixels
        if (seg[5] != 3) return ERR_FMT3;           // Error: Supports only Y/Cb/Cr format
        
        if (autorotate != 0) {
          if (jd->width > jd->height) {
            TFT_Set_Landscape(void);
            DISP_XS=320; DISP_YS=240;}
          else {
            TFT_Set_Portrait(void);
            DISP_XS=240; DISP_YS=320;
          }
        }
        
        // Check three image components
        for (i = 0; i < 3; i++) {
          b = seg[7 + 3 * i];                       // Get sampling factor
          if (!i) {                                 // Y component
            if (b != 0x11 && b != 0x22 && b != 0x21)// Check sampling factor
              return ERR_FMT3;                      // Error: Supports only 4:4:4, 4:2:0 or 4:2:2
            jd->msx = b >> 4; jd->msy = b & 15;     // Size of MCU [blocks]
          } else {                                  // Cb/Cr component
            if (b != 0x11) return ERR_FMT3;         // Error: Sampling factor of Cr/Cb must be 1
          }
          b = seg[8 + 3 * i];                       // Get dequantizer table ID for this component
          if (b > 3) return ERR_FMT3;               // Error: Invalid ID
          jd->qtid[i] = b;
        }
        break;

      case 0xC4:                                    // DHT
                                                    // Load segment data
        if (len > STREAM_BUF) return ERR_MEM2;
        if (jd->virtual_in(jd, seg, len) != len) return ERR_INP;

        // Create Huffman tables
        rc = create_huffman_tbl(jd, seg, len);
        if (rc) return rc;
        break;

      case 0xDB:                                    // DQT
                                                    // Load segment data
        if (len > STREAM_BUF) return ERR_MEM2;
        if (jd->virtual_in(jd, seg, len) != len) return ERR_INP;

        // Create dequantizer tables
        rc = create_qt_tbl(jd, seg, len);
        if (rc) return rc;
        break;

      case 0xDA:                                    // SOS
                                                    // Load segment data
        if (len > STREAM_BUF) return ERR_MEM2;
        if (jd->virtual_in(jd, seg, len) != len) return ERR_INP;
        if (!jd->width || !jd->height) return ERR_FMT1;      // Error: Invalid image size
        if (seg[0] != 3) return ERR_FMT3;                    // Error: Supports only three color components format

        // Check if all tables corresponding to each components have been loaded
        for (i = 0; i < 3; i++) {
          b = seg[2 + 2 * i];                                // Get Huffman table ID
          if (b != 0x00 && b != 0x11) return ERR_FMT3;       // Error: Different table number for DC/AC element
          b = i ? 1 : 0;
          if (!jd->huffbits[b][0] || !jd->huffbits[b][1])    // Check Huffman table for this component
            return ERR_FMT1;                                 // Error: Huffman table not loaded
          if (!jd->qttbl[jd->qtid[i]]) return ERR_FMT1;      // Error: Dequantizer table not loaded
        }
        // Allocate working buffer for MCU and RGB
        n = jd->msy * jd->msx;                               // Number of Y blocks in the MCU
        if (!n) return ERR_FMT1;                             // Error: SOF0 has not been loaded
        len = n * 64 * 2 + 64;                               // Allocate buffer for IDCT and RGB output
        if (len < 256) len = 256;                            // but at least 256 byte is required for IDCT
        jd->workbuf = allocate(jd, len);                     // and it may occupy a part of following MCU working buffer for RGB output
        if (!jd->workbuf) return ERR_MEM1;                   // Error: not enough memory
        jd->mcubuf = allocate(jd, (n + 2) * 64);             // Allocate MCU working buffer
        if (!jd->mcubuf) return ERR_MEM1;                    // Error: not enough memory

        // Preload the JPEG data to extract it from the bit stream
        jd->dptr = seg; jd->dctr = 0; jd->dmsk = 0;          // Prepare to read bit stream
        if (ofs %= STREAM_BUF) {                             // Align read offset to STREAM_BUF
          jd->dctr = jd->virtual_in(jd, seg + ofs, STREAM_BUF - (UINT)ofs);
          jd->dptr = seg + ofs - 1;
        }
        return OK;                         // Initialization succeeded. Ready to decode the JPEG image.

      case 0xC1:        // SOF1             Extended sequential
      case 0xC2:        // SOF2             Progressive
      case 0xC3:        // SOF3             Lossless
      case 0xC5:        // SOF5             Differential sequential
      case 0xC6:        // SOF6             Differential progressive
      case 0xC7:        // SOF7             Differential lossless
      case 0xC9:        // SOF9             Extended sequential, arithmetic coding
      case 0xCA:        // SOF10            Progressive, arithmetic coding
      case 0xCB:        // SOF11            Lossless, arithmetic coding
      case 0xCD:        // SOF13            Differential sequential, arithmetic coding
      case 0xCE:        // SOF14            Differential progressive, arithmetic coding
      case 0xCF:        // SOF15            Differential lossless, arithmetic coding
      case 0xD9:        // EOI
      case 0xDD:        // DRI              Restart interval (not used)
        return ERR_FMT3;                   // Unsuppoted JPEG standard (may be progressive JPEG etc.)
      default:                             // Unknown segment (comment, exif or etc..)
                                           // Skip segment data
        if (jd->virtual_in(jd, 0, len) != len) // Null pointer specifies to skip len bytes
          return ERR_INP;
    }
  }
}


// ----------------------------------
// Start to decode JPEG picture data
// ----------------------------------

static RESULT decode (JPGS* jd, UINT (*virtual_out)(JPGS*, void*, RECT*), BYTE scale) {

  UINT x, y, mx, my;
  WORD rst, rsc;
  RESULT rc;

  if (scale > (autoscale ? 3 : 0)) return ERR_PAR;
  jd->scale = scale;
  mx = jd->msx * 8; my = jd->msy * 8;              // Size of the MCU (pixels)
  jd->dcv[2] = jd->dcv[1] = jd->dcv[0] = 0;        // Initialize DC values
  rst = rsc = 0;
  rc = OK;
  for (y = 0; y < jd->height; y += my) {           // Vertical loop of MCUs
    for (x = 0; x < jd->width; x += mx) {          // Horizontal loop of MCUs
      rc = mcu_load(jd);                           // Load an MCU (decompress Huffman coded data and IDCT)
      if (rc != OK) return rc;
      rc = mcu_output(jd, virtual_out, x, y);      // Output the MCU (color space conversion, scaling and output)
      if (rc != OK) return rc;
    }
  }
  return rc;
}